#include <xtl.h>
#include <ppcintrinsics.h>
#include "kernel.h"
#include "devio.h"
#include "launchInstaller.h"
#include "util.h"

#define MAX_PATCH_SIZE	0x4000
#define PATCH_OFFSET	0x91000

char* consoleTypes[] = {
	"Xenon",
	"Zephyr",
	"Falcon",
	"Jasper",
	"Trinity" // wishful thinking
};

char * PatchNames[] = {
	"GAME:\\files\\patches_xenon.bin",
	"GAME:\\files\\patches_zephyr.bin",
	"GAME:\\files\\patches_falcon.bin",
	"GAME:\\files\\patches_jasper.bin"
};

// reads logical flash
DWORD readFlash(DWORD offset, PBYTE buf, DWORD len)
{
	HANDLE hFile;
	OBJECT_ATTRIBUTES atFlash;
	IO_STATUS_BLOCK ioFlash;
	DWORD bRead = 0, dwPos;
	DWORD sta;
	STRING nFlash = MAKE_STRING("\\Device\\Flash");
	atFlash.RootDirectory = 0;
	atFlash.ObjectName = &nFlash;
	atFlash.Attributes = FILE_ATTRIBUTE_DEVICE;
	sta = NtOpenFile(&hFile, GENERIC_WRITE|GENERIC_READ|SYNCHRONIZE, &atFlash, &ioFlash, OPEN_EXISTING, FILE_SYNCHRONOUS_IO_NONALERT);
	if(sta != 0)
	{
		dprintf("NtOpenFile(flash): %08x (err: %08x)\n", sta, GetLastError());
		return FALSE;
	}

	dwPos = SetFilePointer(hFile, offset, NULL, FILE_BEGIN);
	if(dwPos != INVALID_SET_FILE_POINTER)
	{
		ZeroMemory(buf, len);
		ReadFile(hFile, buf, len, &bRead, NULL);
		NtClose(hFile);
	}
	return bRead;
}

// note while writing flash in this manner that write address must be block aligned
// writes must start at the beginning of a block
// and writes must occur in 1block chunks (or less, but you cannot write multiple times to the same block)
// system takes care of spare (lba/edc) and relocation if necessary
// offset addresses are without spare included
DWORD writeFlash(DWORD offset, PBYTE buf, DWORD len, DWORD writeSize)
{
	HANDLE hFile;
	OBJECT_ATTRIBUTES atFlash;
	IO_STATUS_BLOCK ioFlash;
	DWORD bWrote = 0, tWrote = 0;
	PBYTE src = buf;
	DWORD sta;
	LARGE_INTEGER lOffset;
	STRING nFlash = MAKE_STRING("\\Device\\Flash");
	atFlash.RootDirectory = 0;
	atFlash.ObjectName = &nFlash;
	atFlash.Attributes = FILE_ATTRIBUTE_DEVICE;
	sta = NtOpenFile(&hFile, GENERIC_WRITE|GENERIC_READ|SYNCHRONIZE, &atFlash, &ioFlash, OPEN_EXISTING, FILE_SYNCHRONOUS_IO_NONALERT);
	if(sta != 0)
	{
		dprintf("NtOpenFile(flash): %08x (err: %08x)\n", sta, GetLastError());
		return FALSE;
	}

	lOffset.QuadPart = (LONGLONG)offset;
	while(len > writeSize)
	{
		sta = NtWriteFile(hFile, 0, 0, 0, &ioFlash, src, writeSize, &lOffset);
		if(sta >= 0)
			bWrote = writeSize;
		len -= bWrote;
		src += bWrote;
		lOffset.QuadPart += bWrote;
		tWrote += bWrote;
	}
	if(len > 0)
	{
		sta = NtWriteFile(hFile, 0, 0, 0, &ioFlash, src, len, &lOffset);
		if(sta >= 0)
			bWrote = len;
		tWrote += bWrote;
	}
	NtClose(hFile);
	return tWrote;
}

VOID updatePatches(VOID)
{
	DWORD updateSize, writeSize, blockOff, patchOff, plen, flen;
	BYTE* buf, *patches;
	DWORD fscInfo = __loadvolatilewordbytereverse(0x7feac000,NULL); // read in the flash config word
	// see what console and NAND type we are dealing with
	DWORD conType = (XboxHardwareInfo->Flags >>28)&0xF;
	// enumerate the config word
	switch((fscInfo>>17)&3)
	{
		case 0: // small block flash controller
			switch ((fscInfo >> 4) & 0x3)
			{
				case 1: // 16MB
				case 3: // 64MB
					updateSize = 0x8000; // two blocks
					writeSize = 0x4000;
					blockOff = 0x90000; // start of the two blocks
					patchOff = 0x1000; // start of patches in the two blocks
					break;
				default:
					dprintf("unknown NAND size! (%x)\n", (fscInfo >> 4) & 0x3);
					return;
			}
			break;
		case 1: // big block flash controller
			switch ((fscInfo >> 4) & 0x3)
			{
				case 1: // Small block 16MB setup
					updateSize = 0x8000; // two blocks
					writeSize = 0x4000;
					blockOff = 0x90000; // start of the two blocks
					patchOff = 0x1000; // start of patches in the two blocks
					break;
				case 2: // Large Block: Current Jasper 256MB and 512MB
					updateSize = 0x20000; // one block 128KB
					writeSize = 0x20000;
					blockOff = 0x80000; // start of the block
					patchOff = 0x11000; // start of patches in the block
					break;
				default:
					dprintf("unknown NAND size! (%x)\n", (fscInfo >> 4) & 0x3);
					return;
			}
			break;
		default:
			dprintf("unknown NAND type! (%x)\n", (fscInfo>>17)&3);
			return;
	}
	// read patch data
	buf = new BYTE[MAX_PATCH_SIZE];
	patches = new BYTE[MAX_PATCH_SIZE];
	plen = readFileToBuf(PatchNames[conType], patches, MAX_PATCH_SIZE);
	if(plen != 0)
	{
		flen = readFlash(PATCH_OFFSET, buf, MAX_PATCH_SIZE);
		if(flen != 0)
		{
			// verify they need an update
			if(!byteBufCmp(buf, patches, plen))
			{
				dprintf("\nFreeboot patches included for %s system are not installed\n", consoleTypes[conType]);
				dprintf("Would you like to update them? Press A to update or B to cancel\n");
				if(getConfirm())
				{
					dprintf("Updating patches...\n");
					delete[] buf;
					buf = new BYTE[updateSize];
					// read the block(s)
					flen = readFlash(blockOff, buf, updateSize);
					// clear the patch area so no remnants are left behind
					ZeroMemory(&buf[patchOff], MAX_PATCH_SIZE);
					// inject the patches
					memcpy(&buf[patchOff], patches, plen);
					// write the block(s) back
					writeFlash(blockOff, buf, updateSize, writeSize);
					dprintf("Patch update complete!\n");
				}
				else
					dprintf("Skipping patch update\n");
			}
			else
				dprintf("\nPatches don't need to be updated, already current\n");
		}
		else
			dprintf("reading flash failed!\n");
	}
	else
		dprintf("reading %s to buffer failed!\n", PatchNames[conType]);

	delete[] buf;
	delete[] patches;
}

BOOL patchExists(VOID)
{
	DWORD conType = (XboxHardwareInfo->Flags >>28)&0xF;
	if(fileExists(PatchNames[conType]))
		return TRUE;
	return FALSE;
}

/*
// 16M NAND (hynix)
SFC d4811901 011981d4
0:200	512		userSize
1:10	16		spareSize
2:20	32		pagesPerBlock
3:400	1024	totalBlocks
4:20	32		erasePageSize

// 64M NAND (samsung)
SFC f4811901 011981f4
0:200	512		userSize
1:10	16		spareSize
2:20	32		pagesPerBlock
3:1000	4096	totalBlocks
4:80	128		erasePageSize

// 256M NAND (samsung)
SFC 64308a00 008a3064
0:200	512		userSize
1:10	16		spareSize
2:100	256		pagesPerBlock
3:200	256		totalBlocks
4:20	32		erasePageSize

// 512M NAND
sfc 6430aa00 00aa3064
0:200	512		userSize
1:10	16		spareSize
2:100	256		pagesPerBlock
3:200	512		totalBlocks
4:20	32		erasePageSize
*/

